//--------------------------------------------------------------------------------------
// FroggyAA
// Author: Tristan Lorach
// Email: tlorach@nvidia.com
//
// Implementation of different antialiasing methods.
// - typical MSAA
// - CSAA
// - Hardware AA mixed with FBO for supersampling pass
//   - simple downsampling
//   - downsampling with 1 or 2 kernel filters
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
#include <windows.h> // needed for "LoadLibrary" in GETSVCFACTORY()... need to avoid this ASAP
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include <map>
#include <GL/glew.h>
#include <GL/glut.h>
#include <Cg/CgGL.h>
#include <nvGlutManipulators.h>
#define NV_REPORT_COMPILE_ERRORS
#include <nvShaderUtils.h>
#include <nvGlutWidgets.h>
#include <nvTime.h>

#include "AABox.h"

#ifndef M_PI
#define M_PI 3.141592654f
#endif

float g_curSSSize = 2.0f;
float g_newSSSize = 2.0f;
// the CSAA/MSAA modes depend on the # of color samples and # of coverage samples
// Please see http://developer.nvidia.com/object/coverage-sampled-aa.html for more details
#define MSAA_4x 4, 0, "MSAA_4x"
#define CSAA_8x 4, 8, "CSAA_8x"
#define CSAA_8xQ 8, 8, "CSAA_8xQ"
#define CSAA_16x 4, 16, "CSAA_16x"
#define CSAA_16xQ 8, 16, "CSAA_16xQ"

struct SAAMode 
{ 
	int ds; 
	int cs; 
	const char * name;
};

int g_numAAModes = 5;
SAAMode g_aaModes[] = {
    {MSAA_4x},
    {CSAA_8x},
    {CSAA_8xQ},
    {CSAA_16x},
    {CSAA_16xQ}
};
int					g_curAA = 0;
static AABox*		g_pAABox = NULL;
int					g_mousex = 0, g_mousey = 0;
int					g_width = 0, g_height = 0;
int					g_hfiltersize = 128;
int					g_curTech = 2;
GLuint              g_texid;
nv::GlutUIContext	g_ui;
nv::FrameTimer		g_FrameTimer;

using std::map;

////////////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////////////

#define IMAGE_PATH "../media/textures/"

nv::GlutExamine manipulator;
GLuint fprog, lerp_fprog;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_LERP_LAYERS,
    OPTION_ANIMATE,
    OPTION_DO_AA,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char,UIOption> optionKeyMap;

void key(unsigned char k, int x, int y);
////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////


void init_opengl(int w, int h) 
{
    glewInit();

    if(!glewIsExtensionSupported("GL_NV_framebuffer_multisample_coverage"))
        g_numAAModes = 1; // no using other modes (CSAA)
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDepthMask(true);

    if (!glewIsSupported(
        "GL_VERSION_2_0 "
        "GL_EXT_framebuffer_object "
        "GL_EXT_framebuffer_multisample "
        ))
    {
        printf("Unable to load extensions GL_EXT_framebuffer_object or GL_EXT_frambuffer_multisample \n");
        return;
    }
}

//
// drawing functions
//
//////////////////////////////////////////////////////////////////////

void drawUI()
{
    g_FrameTimer.newFrame();

    g_ui.begin();
    {
		g_ui.beginGroup();
		std::stringstream lText;
		lText << std::string("Frame : ") << ((int) g_FrameTimer.getFramerate()) << std::string(" / ") << (int) (g_FrameTimer.getFrameDuration() * 1000) << std::string(" ms");
		g_ui.doLabel(nv::Rect(),  lText.str().c_str(), 1);
		g_ui.doCheckButton(nv::Rect(), "Wire", &options[OPTION_DISPLAY_WIREFRAME]);
		g_ui.doCheckButton(nv::Rect(), "Animate", &options[OPTION_ANIMATE]);

		{
			g_ui.beginFrame();
			g_ui.doCheckButton(nv::Rect(), "Do SuperSampling AA", &options[OPTION_DO_AA]);
			if(!options[OPTION_DO_AA])
			{
				g_ui.doLabel(nv::Rect(),  "Default MSAA 4x mode from Glut");
			} else
			{

				g_ui.beginFrame(nv::GroupFlags_LayoutHorizontal);
				g_ui.doLabel(nv::Rect(),  "SS factor");
				if(g_ui.doHorizontalSlider(nv::Rect(), 1.0, 2.0, &g_curSSSize))
				{
					delete g_pAABox;
					g_pAABox = new AABox();
					printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
					g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
				}
				g_ui.endFrame();
				std::stringstream lTextFBO;
				lTextFBO << std::string("FBO size : ") 
					<< g_width << std::string("x") << g_height << std::string(" x ") << g_curSSSize
					<< std::string(" = ")
					<< (int)(g_curSSSize*(float)g_width)
					<< std::string(" x ")
					<< (int)(g_curSSSize*(float)g_height);
				g_ui.doLabel(nv::Rect(),  lTextFBO.str().c_str(), 1);

				{
					int newAA = g_curAA;
					static const char* SSAAModes[] = { "4x", "8x", "8xQ", "16x", "16xQ"};
					g_ui.doComboBox( nv::Rect(), 5, SSAAModes, &newAA);
					if(newAA != g_curAA)
					{
						delete g_pAABox;
						g_pAABox = new AABox();
						printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
						g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
					}
				}
				{
					static const char* SSFilterTechniques[] = { "simple bilinear downsampling", "5 taps downsampling", "2 kernels on alpha + 5 taps downsampling" };
					g_ui.doComboBox( nv::Rect(), 3, SSFilterTechniques, &g_curTech);
				}

			}
			g_ui.endFrame();
		}
		g_ui.endGroup();
	}	  
    g_ui.end();
}

void drawSphere(int sphslices )
{
    GLfloat rho, drho, theta, dtheta;
    GLfloat x, y, z;
    GLfloat s, t, ds, dt;
    GLint i, j;
    
    /* texturing: s goes from 0.0/0.25/0.5/0.75/1.0 at +y/+x/-y/-x/+y axis */
    /* t goes from -1.0/+1.0 at z = -radius/+radius (linear along longitudes) */
    /* cannot use triangle fan on texturing (s coord. at top/bottom tip varies) */
    
    int slices = (int)sphslices;
    int stacks = slices;

    ds = 1.0 / sphslices;//slices;
    dt = 1.0 / sphslices;//stacks;
    t = 1.0;
    drho = M_PI / (GLfloat) stacks;
    dtheta = 2.0 * M_PI / (GLfloat) slices;

    //
    // in supersampling mode : we must change the thickness of the lines
    //
    glLineWidth(options[OPTION_DO_AA] ? g_curSSSize : 1.0);
    
    /* draw intermediate stacks as quad strips */
    for (i= 0; i < stacks; i++) {
        rho = i * drho;

        glBegin(GL_LINES);
        s = 0.0;
        for (j = 0; j<=slices; j++) {
            theta = (j==slices) ? 0.0 : j * dtheta;
            x = -sin(theta) * sin(rho);
            z = cos(theta) * sin(rho);
            y = -cos(rho);

            // Let's ask for a larger kernel (alpha = 0) where line are crossing
            glColor4f(0,0,0,fabs(sin(rho)));
            glNormal3f(x, y, z);
            glVertex3f(x, y, z);

            x = -sin(theta) * sin(rho+drho);
            z = cos(theta) * sin(rho+drho);
            y = -cos(rho+drho);

            glNormal3f(x, y, z);
            glVertex3f(x, y, z);
            s += ds;
        }
        glEnd();
        t -= dt;
    }
}

void DrawObjects()
{
	glLoadIdentity();
	manipulator.applyTransform();
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    //glEnable(GL_TEXTURE_2D);
    //glBindTexture(GL_TEXTURE_2D, g_texid);
    glColor4f(0.6, 0.8, 1.0, 1.0);
   	glutSolidTorus(0.1, 0.6, 50, 50);
    glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
   	glutSolidTorus(0.1, 0.6, 50, 50);
    glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
   	glutSolidTorus(0.1, 0.6, 50, 50);
    //glDisable(GL_TEXTURE_2D);
    //glBindTexture(GL_TEXTURE_2D, 0);

    drawSphere(50);

    glRotatef(90.0, 0,1,0);
    glTranslatef(0.4,0.0,0.0);
    glPushMatrix();
    for(int i=0; i < 10; i++)
    {
        // Note : here we put alpha=0, meaning that we will use a larger kernel when in tech '2' (key 3)
	    glColor4f(0.4, 0.9, 0.7, 0.0);
        glutSolidCone(0.01, 0.5, 10, 10);
        glScalef(1.0,1.0,-1.0);
        glutSolidCone(0.01, 0.5, 10, 10);
        glScalef(1.0,1.0,1.0);
        glTranslatef(0.0,0.05,0);
    }
    glPopMatrix();
    glPushMatrix();
    for(int i=0; i < 10; i++)
    {
        // Note : here we put alpha=1, meaning that we will use the normal kernel when tech 3
	    glColor4f(0.7, 0.9, 0.4, 1.0);
        glutSolidCone(0.01, 0.5, 10, 10);
        glScalef(1.0,1.0,-1.0);
        glutSolidCone(0.01, 0.5, 10, 10);
        glScalef(1.0,1.0,1.0);
        glTranslatef(0.0,-0.05,0);
    }
    glPopMatrix();

    glRotatef(-90.0f, 0.0f, 1.0f, 0.0f);
	glColor4f(0.7, 0.5, 0.7, 1.0);
    glTranslatef(0.0,0.0,-0.2);

    glutSolidCone(0.1, 1.0, 10, 10);
    glDisable(GL_LIGHTING);
}
/*------------------------------------------------------

Main rendering functions.

  ------------------------------------------------------*/
void display() 
{
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDepthMask(true);
    glClearColor(0.5, 0.5, 0.5, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);

    // if requested, switch from Backbuffer to offscreen supersampled buffer
    if(options[OPTION_DO_AA])
    {
	    g_pAABox->Activate(); 
    }
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glPolygonMode( GL_FRONT_AND_BACK, options[OPTION_DISPLAY_WIREFRAME] ? GL_LINE : GL_FILL);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
	glDisable(GL_VERTEX_PROGRAM_ARB);
	glPushMatrix();
        DrawObjects();
	glPopMatrix();
    // If requested, downsample (+filter) supersampled offscreen buffer and
    // draw the result with a fullscreen quad
    if(options[OPTION_DO_AA])
    {
	    g_pAABox->Deactivate();
	    g_pAABox->Draw(g_curTech);
    }
	///////////////////////////////////////////////////////////
    glLineWidth(1.0);

	drawUI();

	glutSwapBuffers();
}

//
//
//////////////////////////////////////////////////////////////////////
void idle() {
    if ( options[OPTION_ANIMATE])
        manipulator.idle();
    
    glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void specialkey(int k, int x, int y)
{
    g_ui.specialKeyboard(k, x, y);
    glutPostRedisplay();
}
//
//
//////////////////////////////////////////////////////////////////////
void key(unsigned char k, int x, int y) {
    g_ui.keyboard(k, x, y);

	k = tolower(k);

    if (optionKeyMap.find(k) != optionKeyMap.end())
    {
        options[optionKeyMap[k]] = !options[optionKeyMap[k]];
	}
	
    switch(k) {
        case 13:
            g_curSSSize = g_newSSSize;
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("using AA mode %s\n", g_aaModes[g_curAA].name);
            printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            break;
        case 27:
        case 'q':
            exit(0);
            break;
        //--------------------------------------------------------------------
        // choose one of the 3 techniques
		case '1':
		case '2':
		case '3':
			g_curTech = (int)(k - '1');
			printf("now using technique DownSample_Filter%d\n", g_curTech);
            glutPostRedisplay();
			break;
        //--------------------------------------------------------------------
        // Change the AA mode. See g_aaModes[] and related doc on our web site
        case '=':
        case '+':
            {
            options[OPTION_DO_AA] = true;
            if(g_curAA < (g_numAAModes-1)) g_curAA++;
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("using AA mode %s\n", g_aaModes[g_curAA].name);
            printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;
        case '-':
        case '_':
            {
            options[OPTION_DO_AA] = true;
            if(g_curAA > 0) g_curAA--;
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("using AA mode %s\n", g_aaModes[g_curAA].name);
            printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;
        //--------------------------------------------------------------------
        // Change the resolution of the supersampled buffer
        // default is 2x bigger than the backbuffer. But lower, like 1.5, could
        // still give good results. You can check this thanks to this :
        case '0':
            {
            options[OPTION_DO_AA] = true;
            g_curSSSize += 0.1;
            if(g_curSSSize > 2.0) 
                g_curSSSize = 2.0;
            g_newSSSize = g_curSSSize;
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;
        case '9':
            {
            options[OPTION_DO_AA] = true;
            g_curSSSize -= 0.1;
            if(g_curSSSize < 1.0) 
                g_curSSSize = 1.0f;
            g_newSSSize = g_curSSSize;
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;
        //--------------------------------------------------------------------
        // preset : normal rendering. No AA
        case 'z':
            options[OPTION_DO_AA] = false;
            printf("simple backbuffer. No AA\n");
            break;
        //--------------------------------------------------------------------
        // preset for : no special kernel filtering & no supersampling
        case 'x':
            {
            options[OPTION_DO_AA] = true;
            g_curSSSize = 1.0;
            g_curAA = g_numAAModes==1 ? 0 : 3;
            g_curTech = 0;
            printf("setting for basic %s mode (no special kernel filtering & no supersampling)\n", g_aaModes[g_curAA].name);
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("resizing FBOs : %dx%d\n", g_width, g_height);
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;
        //--------------------------------------------------------------------
        // preset : 
        // supersampling with kernel from technique DownSample_Filter3
        case 'c':
            {
            options[OPTION_DO_AA] = true;
            g_curSSSize = 2.0;
            g_curAA = g_numAAModes==1 ? 0 : 3;
            g_curTech = 2;
            printf("setting for %s mode with special kernel filtering : supersampling with kernel from technique DownSample_Filter3\n", g_aaModes[g_curAA].name);
            delete g_pAABox;
            g_pAABox = new AABox();
            printf("resizing FBOs : %f x %dx%d = %dx%d\n", g_curSSSize, g_width, g_height, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
            g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
            }
            break;

    }
    
	glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void resize(int w, int h) 
{
    g_ui.reshape(w, h);

	g_width = w;
	g_height = h;
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.1, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);

    //
    // Resize FBOs
    // we want to keep consisten ratio with backbuffer.
    // so we need to re-create the AABox
    //
    if(g_pAABox) delete g_pAABox;
    g_pAABox = new AABox();
    printf("using AA mode %s\n", g_aaModes[g_curAA].name);
    printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
    g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);

}

//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y) {

    g_ui.mouse(button, state, glutGetModifiers(), x, y);

	if(!g_ui.isOnFocus()) manipulator.mouse(button, state, x, y);
	glutPostRedisplay();
}

void passiveMotion(int x, int y) {

    g_ui.mouseMotion(x, y);

	g_mousex = x;
	g_mousey = g_height - y;
	glutPostRedisplay();
}
void motion(int x, int y) {

    g_ui.mouseMotion(x, y);

	if(!g_ui.isOnFocus()) manipulator.motion(x, y);
}

void main_menu(int i)
{
  key((unsigned char) i, 0, 0);
}
void aa_menu(int i)
{
    if(i == 10) // No AA
    {
        options[OPTION_DO_AA] = false;
        return;
    }
    options[OPTION_DO_AA] = true;
    g_curAA = i;
    delete g_pAABox;
    g_pAABox = new AABox();
    printf("using AA mode %s\n", g_aaModes[g_curAA].name);
    printf("resizing FBOs : %dx%d x %f = %dx%d\n", g_width, g_height, g_curSSSize, (int)(g_curSSSize*(float)g_width), (int)(g_curSSSize*(float)g_height));
    g_pAABox->Initialize(g_width, g_height, g_curSSSize, g_aaModes[g_curAA].ds, g_aaModes[g_curAA].cs);
}

void init_menus()
{
    int aa_menu_id = glutCreateMenu(aa_menu);
    {
        glutAddMenuEntry("No AA", 10);
        glutAddMenuEntry("MSAA_4x", 0);
        if(g_numAAModes > 1)
        {
            glutAddMenuEntry("CSAA_8x", 1);
            glutAddMenuEntry("CSAA_8xQ", 2);
            glutAddMenuEntry("CSAA_16x", 3);
            glutAddMenuEntry("CSAA_16xQ", 4);
        }
    }
    int tech_menu_id = glutCreateMenu(main_menu);
    {
        glutAddMenuEntry("basic ", '1');
        glutAddMenuEntry("1 kernel filter", '2');
        glutAddMenuEntry("2 kernels based on alpha", '3');
    }
    int object_menu_id = glutCreateMenu(main_menu);
    {
        glutCreateMenu(main_menu);
        glutAddSubMenu("AA mode", aa_menu_id);
        glutAddSubMenu("Downsampling mode", tech_menu_id);
        glutAddMenuEntry("Increment Supersampled buffer [0]", '0');
        glutAddMenuEntry("Decrement Supersampled buffer [9]", '9');
        glutAddMenuEntry("CSAA_8xQ only [c]", 'c');
        glutAddMenuEntry("cool mix : CSAA_8xQ + supersampling 2x + 2 kernel filters", 's');
        glutAddMenuEntry("Quit (esc)", '\033');
        glutAttachMenu(GLUT_RIGHT_BUTTON);
    }
}
//
//
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv) {
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
    glutInitDisplayString("rgb depth double samples=4");
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("Froggy AA");

	init_opengl(800, 600);

    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    manipulator.setDollyPosition( -2.0f);

	glutDisplayFunc(display);
	glutPassiveMotionFunc(passiveMotion);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutReshapeFunc(resize);
    init_menus();
    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = false;

    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;

    optionKeyMap['l'] = OPTION_LERP_LAYERS;
    options[OPTION_LERP_LAYERS] = false;

//    optionKeyMap['z'] = OPTION_DO_AA;
    options[OPTION_DO_AA] = true;

    //print the help info
    printf( "  Commands:\n");
    printf( "  --------\n");
    printf( "  q / [ESC] - Quit the application\n");
    printf( "  [SPACE]   - Toggle continuous animation\n");
    printf( "  f         - Toggle using a fragment program\n");
    printf( "  w         - Toggle displaying wireframe\n");
    printf( "  1 to 3    - change the effect for downsampling\n");
    printf( "  z         - toggles normal backbuffer. No AA\n");
    if(g_numAAModes==1) {
        printf( "  x         - MSAA 4x only (FBO size factor = 1 and no kernel filter used)\n");
        printf( "  c         - MSAA 4x with supersampling (FBO size factor = 2x + kernel filter)\n");
    } else {
        printf( "  x         - CSAA 16x only (FBO size factor = 1 and no kernel filter used)\n");
        printf( "  c         - CSAA 16x with supersampling (FBO size factor = 2x + kernel filter)\n");
    }
    printf( "  -/+       - change the HW Multisampling/CSAA mode for supersampled FBO\n");
    printf( "  9/0       - change the supersampled FBO size factor (default is 2.0, max factor)\n");
    printf( "\n  choose a configuration and press 'a' to alternate and compare\n");

    g_FrameTimer.start();

    glutMainLoop();

    g_FrameTimer.stop();

	if(g_pAABox) delete g_pAABox; // will call Destroy()
	g_pAABox = NULL;
	return 0;
}
